@@ -1,4 +1,4 @@
-# Copyright 2009-2014 MongoDB, Inc.
+# Copyright 2009-2015 MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -20,61 +20,67 @@ The :mod:`gridfs` package is an implementation of GridFS on top of
 .. mongodoc:: gridfs
 """
 
-from gridfs.errors import (NoFile,
-                           UnsupportedAPI)
+from collections import Mapping
+
+from gridfs.errors import NoFile
 from gridfs.grid_file import (GridIn,
                               GridOut,
                               GridOutCursor)
-from pymongo import (MongoClient,
-                     ASCENDING,
+from pymongo import (ASCENDING,
                      DESCENDING)
 from pymongo.database import Database
+from pymongo.errors import ConfigurationError
 
 
 class GridFS(object):
     """An instance of GridFS on top of a single Database.
     """
-    def __init__(self, database, collection="fs", _connect=True):
+    def __init__(self, database, collection="fs"):
         """Create a new instance of :class:`GridFS`.
 
         Raises :class:`TypeError` if `database` is not an instance of
         :class:`~pymongo.database.Database`.
 
+        The `connect` parameter ensures that the underlying
+        :class:`~pymongo.mongo_client.MongoClient` is connected to a server,
+        and creates an index on the "chunks" collection if needed.
+
         :Parameters:
           - `database`: database to use
           - `collection` (optional): root collection to use
+          - `connect` (optional): whether to begin connecting the client in
+            the background
 
-        .. versionadded:: 1.6
-           The `collection` parameter.
+        .. versionchanged:: 3.0
+           `database` must use an acknowledged
+           :attr:`~pymongo.database.Database.write_concern`
 
         .. mongodoc:: gridfs
         """
         if not isinstance(database, Database):
             raise TypeError("database must be an instance of Database")
 
+        if not database.write_concern.acknowledged:
+            raise ConfigurationError('database must use '
+                                     'acknowledged write_concern')
+
         self.__database = database
         self.__collection = database[collection]
         self.__files = self.__collection.files
         self.__chunks = self.__collection.chunks
-        if _connect:
-            self.__ensure_index_files_id()
 
     def __is_secondary(self):
-        client = self.__database.connection
-
-        # Connect the client, so we know if it's connected to the primary.
-        client._ensure_connected()
-        return isinstance(client, MongoClient) and not client.is_primary
+        return not self.__database.client._is_writable()
 
     def __ensure_index_files_id(self):
         if not self.__is_secondary():
-            self.__chunks.ensure_index([("files_id", ASCENDING),
+            self.__chunks.create_index([("files_id", ASCENDING),
                                         ("n", ASCENDING)],
                                        unique=True)
 
     def __ensure_index_filename(self):
         if not self.__is_secondary():
-            self.__files.ensure_index([("filename", ASCENDING),
+            self.__files.create_index([("filename", ASCENDING),
                                        ("uploadDate", DESCENDING)])
 
     def new_file(self, **kwargs):
@@ -90,8 +96,6 @@ class GridFS(object):
 
         :Parameters:
           - `**kwargs` (optional): keyword arguments for file creation
-
-        .. versionadded:: 1.6
         """
         # No need for __ensure_index_files_id() here; GridIn ensures
         # the (files_id, n) index when needed.
@@ -125,24 +129,15 @@ class GridFS(object):
           - `data`: data to be written as a file.
           - `**kwargs` (optional): keyword arguments for file creation
 
-        .. versionadded:: 1.9
-           The ability to write :class:`unicode`, if an `encoding` has
-           been specified as a keyword argument.
-
-        .. versionadded:: 1.6
+        .. versionchanged:: 3.0
+           w=0 writes to GridFS are now prohibited.
         """
         grid_file = GridIn(self.__collection, **kwargs)
-
-        # Start a request - necessary if w=0, harmless otherwise
-        request = self.__collection.database.connection.start_request()
         try:
-            try:
-                grid_file.write(data)
-            finally:
-                grid_file.close()
+            grid_file.write(data)
         finally:
-            # Ensure request is ended even if close() throws error
-            request.end()
+            grid_file.close()
+
         return grid_file._id
 
     def get(self, file_id):
@@ -153,10 +148,12 @@ class GridFS(object):
 
         :Parameters:
           - `file_id`: ``"_id"`` of the file to get
-
-        .. versionadded:: 1.6
         """
-        return GridOut(self.__collection, file_id)
+        gout = GridOut(self.__collection, file_id)
+
+        # Raise NoFile now, instead of on first attribute access.
+        gout._ensure_file()
+        return gout
 
     def get_version(self, filename=None, version=-1, **kwargs):
         """Get a file from GridFS by ``"filename"`` or metadata fields.
@@ -188,12 +185,6 @@ class GridFS(object):
           - `version` (optional): version of the file to get (defaults
             to -1, the most recent version uploaded)
           - `**kwargs` (optional): find files by custom metadata.
-
-        .. versionchanged:: 1.11
-           `filename` defaults to None;
-        .. versionadded:: 1.11
-           Accept keyword arguments to find files by custom metadata.
-        .. versionadded:: 1.9
         """
         self.__ensure_index_filename()
         query = kwargs
@@ -207,7 +198,7 @@ class GridFS(object):
         else:
             cursor.limit(-1).skip(version).sort("uploadDate", ASCENDING)
         try:
-            grid_file = cursor.next()
+            grid_file = next(cursor)
             return GridOut(self.__collection, file_document=grid_file)
         except StopIteration:
             raise NoFile("no version %d for filename %r" % (version, filename))
@@ -222,13 +213,6 @@ class GridFS(object):
         :Parameters:
           - `filename`: ``"filename"`` of the file to get, or `None`
           - `**kwargs` (optional): find files by custom metadata.
-
-        .. versionchanged:: 1.11
-           `filename` defaults to None;
-        .. versionadded:: 1.11
-           Accept keyword arguments to find files by custom metadata. See
-           :meth:`get_version`.
-        .. versionadded:: 1.6
         """
         return self.get_version(filename=filename, **kwargs)
 
@@ -236,7 +220,7 @@ class GridFS(object):
     def delete(self, file_id):
         """Delete a file from GridFS by ``"_id"``.
 
-        Removes all data belonging to the file with ``"_id"``:
+        Deletes all data belonging to the file with ``"_id"``:
         `file_id`.
 
         .. warning:: Any processes/threads reading from the file while
@@ -249,13 +233,10 @@ class GridFS(object):
 
         :Parameters:
           - `file_id`: ``"_id"`` of the file to delete
-
-        .. versionadded:: 1.6
         """
         self.__ensure_index_files_id()
-        self.__files.remove({"_id": file_id},
-                            **self.__files._get_wc_override())
-        self.__chunks.remove({"files_id": file_id})
+        self.__files.delete_many({"_id": file_id})
+        self.__chunks.delete_one({"files_id": file_id})
 
     def list(self):
         """List the names of all files stored in this instance of
@@ -267,9 +248,6 @@ class GridFS(object):
 
         .. versionchanged:: 2.7
            ``list`` ensures an index, the same as ``get_version``.
-
-        .. versionchanged:: 1.6
-           Removed the `collection` argument.
         """
         self.__ensure_index_filename()
 
@@ -279,7 +257,7 @@ class GridFS(object):
             name for name in self.__files.distinct("filename")
             if name is not None]
 
-    def find_one(self, spec_or_id=None, *args, **kwargs):
+    def find_one(self, filter=None, *args, **kwargs):
         """Get a single file from gridfs.
 
         All arguments to :meth:`find` are also valid arguments for
@@ -290,18 +268,18 @@ class GridFS(object):
             file = fs.find_one({"filename": "lisa.txt"})
 
         :Parameters:
-          - `spec_or_id` (optional): a dictionary specifying
-            the query to be performing OR any other type to be used  as
+          - `filter` (optional): a dictionary specifying
+            the query to be performing OR any other type to be used as
             the value for a query for ``"_id"`` in the file collection.
           - `*args` (optional): any additional positional arguments are
             the same as the arguments to :meth:`find`.
           - `**kwargs` (optional): any additional keyword arguments
             are the same as the arguments to :meth:`find`.
         """
-        if spec_or_id is not None and not isinstance(spec_or_id, dict):
-            spec_or_id = {"_id": spec_or_id}
+        if filter is not None and not isinstance(filter, Mapping):
+            filter = {"_id": filter}
 
-        for f in self.find(spec_or_id, *args, **kwargs):
+        for f in self.find(filter, *args, **kwargs):
             return f
 
         return None
@@ -313,12 +291,14 @@ class GridFS(object):
         arbitrary queries on the files collection. Can be combined
         with other modifiers for additional control. For example::
 
-          for grid_out in fs.find({"filename": "lisa.txt"}, timeout=False):
+          for grid_out in fs.find({"filename": "lisa.txt"},
+                                  no_cursor_timeout=True):
               data = grid_out.read()
 
         would iterate through all versions of "lisa.txt" stored in GridFS.
-        Note that setting timeout to False may be important to prevent the
-        cursor from timing out during long multi-file processing work.
+        Note that setting no_cursor_timeout to True may be important to
+        prevent the cursor from timing out during long multi-file processing
+        work.
 
         As another example, the call::
 
@@ -332,43 +312,32 @@ class GridFS(object):
         in :class:`~pymongo.collection.Collection`.
 
         :Parameters:
-          - `spec` (optional): a SON object specifying elements which
+          - `filter` (optional): a SON object specifying elements which
             must be present for a document to be included in the
             result set
           - `skip` (optional): the number of files to omit (from
             the start of the result set) when returning the results
           - `limit` (optional): the maximum number of results to
             return
-          - `timeout` (optional): if True (the default), any returned
-            cursor is closed by the server after 10 minutes of
-            inactivity. If set to False, the returned cursor will never
+          - `no_cursor_timeout` (optional): if False (the default), any
+            returned cursor is closed by the server after 10 minutes of
+            inactivity. If set to True, the returned cursor will never
             time out on the server. Care should be taken to ensure that
-            cursors with timeout turned off are properly closed.
+            cursors with no_cursor_timeout turned on are properly closed.
           - `sort` (optional): a list of (key, direction) pairs
             specifying the sort order for this query. See
             :meth:`~pymongo.cursor.Cursor.sort` for details.
-          - `max_scan` (optional): limit the number of file documents
-            examined when performing the query
-          - `read_preference` (optional): The read preference for
-            this query.
-          - `tag_sets` (optional): The tag sets for this query.
-          - `secondary_acceptable_latency_ms` (optional): Any replica-set
-            member whose ping time is within secondary_acceptable_latency_ms of
-            the nearest member may accept reads. Default 15 milliseconds.
-            **Ignored by mongos** and must be configured on the command line.
-            See the localThreshold_ option for more information.
-          - `compile_re` (optional): if ``False``, don't attempt to compile
-            BSON regex objects into Python regexes. Return instances of
-            :class:`~bson.regex.Regex` instead.
 
         Raises :class:`TypeError` if any of the arguments are of
         improper type. Returns an instance of
         :class:`~gridfs.grid_file.GridOutCursor`
         corresponding to this query.
 
+        .. versionchanged:: 3.0
+           Removed the read_preference, tag_sets, and
+           secondary_acceptable_latency_ms options.
         .. versionadded:: 2.7
         .. mongodoc:: find
-        .. _localThreshold: http://docs.mongodb.org/manual/reference/mongos/#cmdoption-mongos--localThreshold
         """
         return GridOutCursor(self.__collection, *args, **kwargs)
 
@@ -404,26 +373,7 @@ class GridFS(object):
             document to check for
           - `**kwargs` (optional): keyword arguments are used as a
             query document, if they're present.
-
-        .. versionadded:: 1.8
         """
         if kwargs:
             return self.__files.find_one(kwargs, ["_id"]) is not None
         return self.__files.find_one(document_or_id, ["_id"]) is not None
-
-    def open(self, *args, **kwargs):
-        """No longer supported.
-
-        .. versionchanged:: 1.6
-           The open method is no longer supported.
-        """
-        raise UnsupportedAPI("The open method is no longer supported.")
-
-    def remove(self, *args, **kwargs):
-        """No longer supported.
-
-        .. versionchanged:: 1.6
-           The remove method is no longer supported.
-        """
-        raise UnsupportedAPI("The remove method is no longer supported. "
-                             "Please use the delete method instead.")
