@@ -1,4 +1,4 @@
-# Copyright 2009-2014 MongoDB, Inc.
+# Copyright 2009-2015 MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -20,16 +20,14 @@ import os
 
 from bson.binary import Binary
 from bson.objectid import ObjectId
-from bson.py3compat import (b, binary_type, next_item,
-                            string_types, text_type, StringIO)
+from bson.py3compat import text_type, StringIO
 from gridfs.errors import (CorruptGridFile,
                            FileExists,
-                           NoFile,
-                           UnsupportedAPI)
+                           NoFile)
 from pymongo import ASCENDING
 from pymongo.collection import Collection
 from pymongo.cursor import Cursor
-from pymongo.errors import DuplicateKeyError
+from pymongo.errors import ConfigurationError, DuplicateKeyError
 from pymongo.read_preferences import ReadPreference
 
 try:
@@ -42,18 +40,17 @@ except AttributeError:
     _SEEK_CUR = 1
     _SEEK_END = 2
 
-EMPTY = b("")
-NEWLN = b("\n")
+EMPTY = b""
+NEWLN = b"\n"
 
 """Default chunk size, in bytes."""
 # Slightly under a power of 2, to work well with server's record allocations.
 DEFAULT_CHUNK_SIZE = 255 * 1024
 
 
-def _create_property(field_name, docstring,
-                      read_only=False, closed_only=False):
-    """Helper for creating properties to read/write to files.
-    """
+def _grid_in_property(field_name, docstring, read_only=False,
+                      closed_only=False):
+    """Create a GridIn property."""
     def getter(self):
         if closed_only and not self._closed:
             raise AttributeError("can only get %r on a closed file" %
@@ -65,13 +62,12 @@ def _create_property(field_name, docstring,
 
     def setter(self, value):
         if self._closed:
-            self._coll.files.update({"_id": self._file["_id"]},
-                                    {"$set": {field_name: value}},
-                                    **self._coll._get_wc_override())
+            self._coll.files.update_one({"_id": self._file["_id"]},
+                                        {"$set": {field_name: value}})
         self._file[field_name] = value
 
     if read_only:
-        docstring = docstring + "\n\nThis attribute is read-only."
+        docstring += "\n\nThis attribute is read-only."
     elif closed_only:
         docstring = "%s\n\n%s" % (docstring, "This attribute is read-only and "
                                   "can only be read after :meth:`close` "
@@ -82,6 +78,20 @@ def _create_property(field_name, docstring,
     return property(getter, doc=docstring)
 
 
+def _grid_out_property(field_name, docstring):
+    """Create a GridOut property."""
+    def getter(self):
+        self._ensure_file()
+
+        # Protect against PHP-237
+        if field_name == 'length':
+            return self._file.get(field_name, 0)
+        return self._file.get(field_name, None)
+
+    docstring += "\n\nThis attribute is read-only."
+    return property(getter, doc=docstring)
+
+
 class GridIn(object):
     """Class to write data to GridFS.
     """
@@ -119,34 +129,23 @@ class GridIn(object):
             that is written to the file will be converted to
             :class:`bytes`.
 
-        If you turn off write-acknowledgment for performance reasons, it is
-        critical to wrap calls to :meth:`write` and :meth:`close` within a
-        single request:
-
-           >>> from pymongo import MongoClient
-           >>> from gridfs import GridFS
-           >>> client = MongoClient(w=0) # turn off write acknowledgment
-           >>> fs = GridFS(client.database)
-           >>> gridin = fs.new_file()
-           >>> request = client.start_request()
-           >>> try:
-           ...     for i in range(10):
-           ...         gridin.write('foo')
-           ...     gridin.close()
-           ... finally:
-           ...     request.end()
-
-        In Python 2.5 and later this code can be simplified with a
-        with-statement, see :doc:`/examples/requests` for more information.
-
         :Parameters:
           - `root_collection`: root collection to write to
           - `**kwargs` (optional): file level options (see above)
+
+        .. versionchanged:: 3.0
+           `root_collection` must use an acknowledged
+           :attr:`~pymongo.collection.Collection.write_concern`
         """
         if not isinstance(root_collection, Collection):
             raise TypeError("root_collection must be an "
                             "instance of Collection")
 
+        # With w=0, 'filemd5' might run before the final chunks are written.
+        if not root_collection.write_concern.acknowledged:
+            raise ConfigurationError('root_collection must use '
+                                     'acknowledged write_concern')
+
         # Handle alternative naming
         if "content_type" in kwargs:
             kwargs["contentType"] = kwargs.pop("content_type")
@@ -167,7 +166,7 @@ class GridIn(object):
 
     def _ensure_index(self):
         if not object.__getattribute__(self, "_ensured_index"):
-            self._coll.chunks.ensure_index(
+            self._coll.chunks.create_index(
                 [("files_id", ASCENDING), ("n", ASCENDING)],
                 unique=True)
             object.__setattr__(self, "_ensured_index", True)
@@ -178,19 +177,19 @@ class GridIn(object):
         """
         return self._closed
 
-    _id = _create_property("_id", "The ``'_id'`` value for this file.",
+    _id = _grid_in_property("_id", "The ``'_id'`` value for this file.",
                             read_only=True)
-    filename = _create_property("filename", "Name of this file.")
-    name = _create_property("filename", "Alias for `filename`.")
-    content_type = _create_property("contentType", "Mime-type for this file.")
-    length = _create_property("length", "Length (in bytes) of this file.",
+    filename = _grid_in_property("filename", "Name of this file.")
+    name = _grid_in_property("filename", "Alias for `filename`.")
+    content_type = _grid_in_property("contentType", "Mime-type for this file.")
+    length = _grid_in_property("length", "Length (in bytes) of this file.",
                                closed_only=True)
-    chunk_size = _create_property("chunkSize", "Chunk size for this file.",
+    chunk_size = _grid_in_property("chunkSize", "Chunk size for this file.",
                                    read_only=True)
-    upload_date = _create_property("uploadDate",
+    upload_date = _grid_in_property("uploadDate",
                                     "Date that this file was uploaded.",
                                     closed_only=True)
-    md5 = _create_property("md5", "MD5 of the contents of this file "
+    md5 = _grid_in_property("md5", "MD5 of the contents of this file "
                             "(generated on the server).",
                             closed_only=True)
 
@@ -210,9 +209,8 @@ class GridIn(object):
             # them now.
             self._file[name] = value
             if self._closed:
-                self._coll.files.update({"_id": self._file["_id"]},
-                                        {"$set": {name: value}},
-                                        **self._coll._get_wc_override())
+                self._coll.files.update_one({"_id": self._file["_id"]},
+                                            {"$set": {name: value}})
 
     def __flush_data(self, data):
         """Flush `data` to a chunk.
@@ -230,7 +228,7 @@ class GridIn(object):
                  "data": Binary(data)}
 
         try:
-            self._chunks.insert(chunk)
+            self._chunks.insert_one(chunk)
         except DuplicateKeyError:
             self._raise_file_exists(self._file['_id'])
         self._chunk_number += 1
@@ -250,14 +248,6 @@ class GridIn(object):
             self.__flush_buffer()
 
             db = self._coll.database
-
-            # See PYTHON-417, "Sharded GridFS fails with exception: chunks out
-            # of order." Inserts via mongos, even if they use a single
-            # connection, can succeed out-of-order due to the writebackListener.
-            # We mustn't call "filemd5" until all inserts are complete, which
-            # we ensure by calling getLastError (and ignoring the result).
-            db.command('getlasterror', read_preference=ReadPreference.PRIMARY)
-
             md5 = db.command(
                 "filemd5", self._id, root=self._coll.name,
                 read_preference=ReadPreference.PRIMARY)["md5"]
@@ -266,8 +256,7 @@ class GridIn(object):
             self._file["length"] = self._position
             self._file["uploadDate"] = datetime.datetime.utcnow()
 
-            return self._coll.files.insert(self._file,
-                                           **self._coll._get_wc_override())
+            return self._coll.files.insert_one(self._file)
         except DuplicateKeyError:
             self._raise_file_exists(self._id)
 
@@ -306,10 +295,6 @@ class GridIn(object):
         :Parameters:
           - `data`: string of bytes or file-like object to be written
             to the file
-
-        .. versionadded:: 1.9
-           The ability to write :class:`unicode`, if the file has an
-           :attr:`encoding` attribute.
         """
         if self._closed:
             raise ValueError("cannot write to a closed file")
@@ -319,9 +304,9 @@ class GridIn(object):
             read = data.read
         except AttributeError:
             # string
-            if not isinstance(data, string_types):
+            if not isinstance(data, (text_type, bytes)):
                 raise TypeError("can only write strings or file-like objects")
-            if isinstance(data, unicode):
+            if isinstance(data, text_type):
                 try:
                     data = data.encode(self.encoding)
                 except AttributeError:
@@ -371,8 +356,7 @@ class GridIn(object):
 class GridOut(object):
     """Class to read data out of GridFS.
     """
-    def __init__(self, root_collection, file_id=None, file_document=None,
-                 _connect=True):
+    def __init__(self, root_collection, file_id=None, file_document=None):
         """Read a file from GridFS
 
         Application developers should generally not need to
@@ -386,11 +370,13 @@ class GridOut(object):
 
         :Parameters:
           - `root_collection`: root collection to read from
-          - `file_id`: value of ``"_id"`` for the file to read
-          - `file_document`: file document from `root_collection.files`
+          - `file_id` (optional): value of ``"_id"`` for the file to read
+          - `file_document` (optional): file document from
+            `root_collection.files`
 
-        .. versionadded:: 1.9
-           The `file_document` parameter.
+        .. versionchanged:: 3.0
+           Creating a GridOut does not immediately retrieve the file metadata
+           from the server. Metadata is fetched when first needed.
         """
         if not isinstance(root_collection, Collection):
             raise TypeError("root_collection must be an "
@@ -402,27 +388,19 @@ class GridOut(object):
         self.__buffer = EMPTY
         self.__position = 0
         self._file = file_document
-        if _connect:
-            self._ensure_file()
 
-    _id = _create_property("_id", "The ``'_id'`` value for this file.", True)
-    filename = _create_property("filename", "Name of this file.", True)
-    name = _create_property("filename", "Alias for `filename`.", True)
-    content_type = _create_property("contentType", "Mime-type for this file.",
-                                     True)
-    length = _create_property("length", "Length (in bytes) of this file.",
-                               True)
-    chunk_size = _create_property("chunkSize", "Chunk size for this file.",
-                                   True)
-    upload_date = _create_property("uploadDate",
-                                    "Date that this file was first uploaded.",
-                                    True)
-    aliases = _create_property("aliases", "List of aliases for this file.",
-                                True)
-    metadata = _create_property("metadata", "Metadata attached to this file.",
-                                 True)
-    md5 = _create_property("md5", "MD5 of the contents of this file "
-                            "(generated on the server).", True)
+    _id = _grid_out_property("_id", "The ``'_id'`` value for this file.")
+    filename = _grid_out_property("filename", "Name of this file.")
+    name = _grid_out_property("filename", "Alias for `filename`.")
+    content_type = _grid_out_property("contentType", "Mime-type for this file.")
+    length = _grid_out_property("length", "Length (in bytes) of this file.")
+    chunk_size = _grid_out_property("chunkSize", "Chunk size for this file.")
+    upload_date = _grid_out_property("uploadDate",
+                                     "Date that this file was first uploaded.")
+    aliases = _grid_out_property("aliases", "List of aliases for this file.")
+    metadata = _grid_out_property("metadata", "Metadata attached to this file.")
+    md5 = _grid_out_property("md5", "MD5 of the contents of this file "
+                             "(generated on the server).")
 
     def _ensure_file(self):
         if not self._file:
@@ -441,7 +419,6 @@ class GridOut(object):
         """Reads a chunk at a time. If the current position is within a
         chunk the remainder of the chunk is returned.
         """
-        self._ensure_file()
         received = len(self.__buffer)
         chunk_data = EMPTY
         chunk_size = int(self.chunk_size)
@@ -500,11 +477,9 @@ class GridOut(object):
 
         :Parameters:
          - `size` (optional): the maximum number of bytes to read
-
-        .. versionadded:: 1.9
         """
         if size == 0:
-            return b('')
+            return b''
 
         remainder = int(self.length) - self.__position
         if size < 0 or size > remainder:
@@ -609,28 +584,17 @@ class GridOutIterator(object):
         if not chunk:
             raise CorruptGridFile("no chunk #%d" % self.__current_chunk)
         self.__current_chunk += 1
-        return binary_type(chunk["data"])
+        return bytes(chunk["data"])
 
-
-class GridFile(object):
-    """No longer supported.
-
-    .. versionchanged:: 1.6
-       The GridFile class is no longer supported.
-    """
-    def __init__(self, *args, **kwargs):
-        raise UnsupportedAPI("The GridFile class is no longer supported. "
-                             "Please use GridIn or GridOut instead.")
+    __next__ = next
 
 
 class GridOutCursor(Cursor):
     """A cursor / iterator for returning GridOut objects as the result
     of an arbitrary query against the GridFS files collection.
     """
-    def __init__(self, collection, spec=None, skip=0, limit=0,
-                 timeout=True, sort=None, max_scan=None,
-                 read_preference=None, tag_sets=None,
-                 secondary_acceptable_latency_ms=None, compile_re=True):
+    def __init__(self, collection, filter=None, skip=0, limit=0,
+                 no_cursor_timeout=False, sort=None):
         """Create a new cursor, similar to the normal
         :class:`~pymongo.cursor.Cursor`.
 
@@ -644,25 +608,19 @@ class GridOutCursor(Cursor):
         # Hold on to the base "fs" collection to create GridOut objects later.
         self.__root_collection = collection
 
-        # Copy these settings from collection if they are not set by caller.
-        read_preference = read_preference or collection.files.read_preference
-        tag_sets = tag_sets or collection.files.tag_sets
-        latency = (secondary_acceptable_latency_ms
-                   or collection.files.secondary_acceptable_latency_ms)
-
         super(GridOutCursor, self).__init__(
-            collection.files, spec, skip=skip, limit=limit, timeout=timeout,
-            sort=sort, max_scan=max_scan, read_preference=read_preference,
-            secondary_acceptable_latency_ms=latency, compile_re=compile_re,
-            tag_sets=tag_sets)
+            collection.files, filter, skip=skip, limit=limit,
+            no_cursor_timeout=no_cursor_timeout, sort=sort)
 
     def next(self):
         """Get next GridOut object from cursor.
         """
         # Work around "super is not iterable" issue in Python 3.x
-        next_file = getattr(super(GridOutCursor, self), next_item)()
+        next_file = super(GridOutCursor, self).next()
         return GridOut(self.__root_collection, file_document=next_file)
 
+    __next__ = next
+
     def add_option(self, *args, **kwargs):
         raise NotImplementedError("Method does not exist for GridOutCursor")
 
