@@ -1,4 +1,4 @@
-# Copyright 2014 MongoDB, Inc.
+# Copyright 2014-2015 MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -16,28 +16,23 @@
 
 from collections import deque
 
-from pymongo import helpers, message
-from pymongo.errors import AutoReconnect, CursorNotFound
+from bson.py3compat import integer_types
+from pymongo import helpers
+from pymongo.errors import AutoReconnect, CursorNotFound, NotMasterError
+from pymongo.message import _GetMore
 
 
 class CommandCursor(object):
     """A cursor / iterator over command cursors.
     """
 
-    def __init__(self, collection, cursor_info,
-                 conn_id, compile_re=True, retrieved=0):
+    def __init__(self, collection, cursor_info, address, retrieved=0):
         """Create a new command cursor.
         """
         self.__collection = collection
         self.__id = cursor_info['id']
-        self.__conn_id = conn_id
+        self.__address = address
         self.__data = deque(cursor_info['firstBatch'])
-        self.__decode_opts = (
-            collection.database.connection.document_class,
-            collection.database.connection.tz_aware,
-            collection.uuid_subtype,
-            compile_re
-        )
         self.__retrieved = retrieved
         self.__batch_size = 0
         self.__killed = False
@@ -55,11 +50,8 @@ class CommandCursor(object):
         """Closes this cursor.
         """
         if self.__id and not self.__killed:
-            client = self.__collection.database.connection
-            if self.__conn_id is not None:
-                client.close_cursor(self.__id, self.__conn_id)
-            else:
-                client.close_cursor(self.__id)
+            self.__collection.database.client.close_cursor(self.__id,
+                                                           self.__address)
         self.__killed = True
 
     def close(self):
@@ -85,7 +77,7 @@ class CommandCursor(object):
         :Parameters:
           - `batch_size`: The size of each batch of results requested.
         """
-        if not isinstance(batch_size, (int, long)):
+        if not isinstance(batch_size, integer_types):
             raise TypeError("batch_size must be an integer")
         if batch_size < 0:
             raise ValueError("batch_size must be >= 0")
@@ -93,14 +85,13 @@ class CommandCursor(object):
         self.__batch_size = batch_size == 1 and 2 or batch_size
         return self
 
-    def __send_message(self, msg):
+    def __send_message(self, operation):
         """Send a getmore message and handle the response.
         """
-        client = self.__collection.database.connection
+        client = self.__collection.database.client
         try:
-            res = client._send_message_with_response(
-                msg, _connection_to_use=self.__conn_id)
-            self.__conn_id, (response, dummy0, dummy1) = res
+            response = client._send_message_with_response(
+                operation, address=self.__address)
         except AutoReconnect:
             # Don't try to send kill cursors on another socket
             # or to another server. It can cause a _pinValue
@@ -110,26 +101,26 @@ class CommandCursor(object):
             raise
 
         try:
-            response = helpers._unpack_response(response,
-                                                self.__id,
-                                                *self.__decode_opts)
+            doc = helpers._unpack_response(response.data,
+                                           self.__id,
+                                           self.__collection.codec_options)
         except CursorNotFound:
             self.__killed = True
             raise
-        except AutoReconnect:
+        except NotMasterError:
             # Don't send kill cursors to another server after a "not master"
             # error. It's completely pointless.
             self.__killed = True
-            client.disconnect()
+            client._reset_server_and_request_check(self.address)
             raise
-        self.__id = response["cursor_id"]
+        self.__id = doc["cursor_id"]
 
-        assert response["starting_from"] == self.__retrieved, (
+        assert doc["starting_from"] == self.__retrieved, (
             "Result batch started from %s, expected %s" % (
-                response['starting_from'], self.__retrieved))
+                doc['starting_from'], self.__retrieved))
 
-        self.__retrieved += response["number_returned"]
-        self.__data = deque(response["data"])
+        self.__retrieved += doc["number_returned"]
+        self.__data = deque(doc["data"])
 
     def _refresh(self):
         """Refreshes the cursor with more data from the server.
@@ -143,8 +134,7 @@ class CommandCursor(object):
 
         if self.__id:  # Get More
             self.__send_message(
-                message.get_more(self.__ns,
-                                 self.__batch_size, self.__id))
+                _GetMore(self.__ns, self.__batch_size, self.__id))
 
         else:  # Cursor id is zero nothing else to return
             self.__killed = True
@@ -161,6 +151,14 @@ class CommandCursor(object):
         """Returns the id of the cursor."""
         return self.__id
 
+    @property
+    def address(self):
+        """The (host, port) of the server used, or None.
+
+        .. versionadded:: 3.0
+        """
+        return self.__address
+
     def __iter__(self):
         return self
 
@@ -173,6 +171,8 @@ class CommandCursor(object):
         else:
             raise StopIteration
 
+    __next__ = next
+
     def __enter__(self):
         return self
 
