@@ -1,4 +1,4 @@
-# Copyright 2009-2014 MongoDB, Inc.
+# Copyright 2009-2015 MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -13,17 +13,25 @@
 # limitations under the License.
 
 """Cursor class to iterate over Mongo query results."""
+
 import copy
 from collections import deque
 
 from bson import RE_TYPE
 from bson.code import Code
+from bson.py3compat import (iteritems,
+                            integer_types,
+                            string_type)
 from bson.son import SON
-from pymongo import helpers, message, read_preferences
-from pymongo.read_preferences import ReadPreference, secondary_ok_commands
+from pymongo import helpers
+from pymongo.common import validate_boolean, validate_is_mapping
 from pymongo.errors import (AutoReconnect,
+                            ConnectionFailure,
                             InvalidOperation,
+                            NotMasterError,
                             OperationFailure)
+from pymongo.message import _GetMore, _Query
+from pymongo.read_preferences import ReadPreference
 
 _QUERY_OPTIONS = {
     "tailable_cursor": 2,
@@ -35,6 +43,35 @@ _QUERY_OPTIONS = {
     "partial": 128}
 
 
+class CursorType(object):
+    NON_TAILABLE = 0
+    """The standard cursor type."""
+
+    TAILABLE = _QUERY_OPTIONS["tailable_cursor"]
+    """The tailable cursor type.
+
+    Tailable cursors are only for use with capped collections. They are not
+    closed when the last data is retrieved but are kept open and the cursor
+    location marks the final document position. If more data is received
+    iteration of the cursor will continue from the last document received.
+    """
+
+    TAILABLE_AWAIT = TAILABLE | _QUERY_OPTIONS["await_data"]
+    """A tailable cursor with the await option set.
+
+    Creates a tailable cursor that will wait for a few seconds after returning
+    the full result set so that it can capture and return additional data added
+    during the query.
+    """
+
+    EXHAUST = _QUERY_OPTIONS["exhaust"]
+    """An exhaust cursor.
+
+    MongoDB will stream batched results to the client without waiting for the
+    client to request each batch, reducing latency.
+    """
+
+
 # This has to be an old style class due to
 # http://bugs.jython.org/issue1057
 class _SocketManager:
@@ -53,34 +90,19 @@ class _SocketManager:
         """
         if not self.__closed:
             self.__closed = True
-            self.pool.maybe_return_socket(self.sock)
+            self.pool.return_socket(self.sock)
             self.sock, self.pool = None, None
 
-    def error(self):
-        """Clean up after an error on the managed socket.
-        """
-        if self.sock:
-            self.sock.close()
 
-        # Return the closed socket to avoid a semaphore leak in the pool.
-        self.close()
-
-
-# TODO might be cool to be able to do find().include("foo") or
-# find().exclude(["bar", "baz"]) or find().slice("a", 1, 2) as an
-# alternative to the fields specifier.
 class Cursor(object):
     """A cursor / iterator over Mongo query results.
     """
 
-    def __init__(self, collection, spec=None, fields=None, skip=0, limit=0,
-                 timeout=True, snapshot=False, tailable=False, sort=None,
-                 max_scan=None, as_class=None, slave_okay=False,
-                 await_data=False, partial=False, manipulate=True,
-                 read_preference=ReadPreference.PRIMARY,
-                 tag_sets=[{}], secondary_acceptable_latency_ms=None,
-                 exhaust=False, compile_re=True, _must_use_master=False,
-                 _uuid_subtype=None, **kwargs):
+    def __init__(self, collection, filter=None, projection=None, skip=0,
+                 limit=0, no_cursor_timeout=False,
+                 cursor_type=CursorType.NON_TAILABLE,
+                 sort=None, allow_partial_results=False, oplog_replay=False,
+                 modifiers=None, batch_size=0, manipulate=True):
         """Create a new cursor.
 
         Should not be called directly by application developers - see
@@ -90,57 +112,60 @@ class Cursor(object):
         """
         self.__id = None
 
+        spec = filter
         if spec is None:
             spec = {}
 
-        if not isinstance(spec, dict):
-            raise TypeError("spec must be an instance of dict")
+        validate_is_mapping("filter", spec)
         if not isinstance(skip, int):
             raise TypeError("skip must be an instance of int")
         if not isinstance(limit, int):
             raise TypeError("limit must be an instance of int")
-        if not isinstance(timeout, bool):
-            raise TypeError("timeout must be an instance of bool")
-        if not isinstance(snapshot, bool):
-            raise TypeError("snapshot must be an instance of bool")
-        if not isinstance(tailable, bool):
-            raise TypeError("tailable must be an instance of bool")
-        if not isinstance(slave_okay, bool):
-            raise TypeError("slave_okay must be an instance of bool")
-        if not isinstance(await_data, bool):
-            raise TypeError("await_data must be an instance of bool")
-        if not isinstance(partial, bool):
-            raise TypeError("partial must be an instance of bool")
-        if not isinstance(exhaust, bool):
-            raise TypeError("exhaust must be an instance of bool")
+        validate_boolean("no_cursor_timeout", no_cursor_timeout)
+        if cursor_type not in (CursorType.NON_TAILABLE, CursorType.TAILABLE,
+                               CursorType.TAILABLE_AWAIT, CursorType.EXHAUST):
+            raise ValueError("not a valid value for cursor_type")
+        validate_boolean("allow_partial_results", allow_partial_results)
+        validate_boolean("oplog_replay", oplog_replay)
+        if modifiers is not None:
+            validate_is_mapping("modifiers", modifiers)
+        if not isinstance(batch_size, integer_types):
+            raise TypeError("batch_size must be an integer")
+        if batch_size < 0:
+            raise ValueError("batch_size must be >= 0")
 
-        if fields is not None:
-            if not fields:
-                fields = {"_id": 1}
-            if not isinstance(fields, dict):
-                fields = helpers._fields_list_to_dict(fields)
-
-        if as_class is None:
-            as_class = collection.database.connection.document_class
+        if projection is not None:
+            if not projection:
+                projection = {"_id": 1}
+            projection = helpers._fields_list_to_dict(projection, "projection")
 
         self.__collection = collection
         self.__spec = spec
-        self.__fields = fields
+        self.__projection = projection
         self.__skip = skip
         self.__limit = limit
+        self.__batch_size = batch_size
+        self.__modifiers = modifiers and modifiers.copy() or {}
+        self.__ordering = sort and helpers._index_document(sort) or None
+        self.__max_scan = None
+        self.__explain = False
+        self.__hint = None
+        self.__comment = None
         self.__max_time_ms = None
-        self.__batch_size = 0
         self.__max = None
         self.__min = None
+        self.__manipulate = manipulate
 
         # Exhaust cursor support
-        if self.__collection.database.connection.is_mongos and exhaust:
-            raise InvalidOperation('Exhaust cursors are '
-                                   'not supported by mongos')
-        if limit and exhaust:
-            raise InvalidOperation("Can't use limit and exhaust together.")
-        self.__exhaust = exhaust
+        self.__exhaust = False
         self.__exhaust_mgr = None
+        if cursor_type == CursorType.EXHAUST:
+            if self.__collection.database.client.is_mongos:
+                raise InvalidOperation('Exhaust cursors are '
+                                       'not supported by mongos')
+            if limit:
+                raise InvalidOperation("Can't use limit and exhaust together.")
+            self.__exhaust = True
 
         # This is ugly. People want to be able to do cursor[5:5] and
         # get an empty result set (old behavior was an
@@ -150,63 +175,31 @@ class Cursor(object):
         # it anytime we change __limit.
         self.__empty = False
 
-        self.__snapshot = snapshot
-        self.__ordering = sort and helpers._index_document(sort) or None
-        self.__max_scan = max_scan
-        self.__explain = False
-        self.__hint = None
-        self.__comment = None
-        self.__as_class = as_class
-        self.__slave_okay = slave_okay
-        self.__manipulate = manipulate
-        self.__read_preference = read_preference
-        self.__tag_sets = tag_sets
-        self.__secondary_acceptable_latency_ms = secondary_acceptable_latency_ms
-        self.__tz_aware = collection.database.connection.tz_aware
-        self.__compile_re = compile_re
-        self.__must_use_master = _must_use_master
-        self.__uuid_subtype = _uuid_subtype or collection.uuid_subtype
-
         self.__data = deque()
-        self.__connection_id = None
+        self.__address = None
         self.__retrieved = 0
         self.__killed = False
 
-        self.__query_flags = 0
-        if tailable:
-            self.__query_flags |= _QUERY_OPTIONS["tailable_cursor"]
-        if not timeout:
-            self.__query_flags |= _QUERY_OPTIONS["no_timeout"]
-        if tailable and await_data:
-            self.__query_flags |= _QUERY_OPTIONS["await_data"]
-        if exhaust:
-            self.__query_flags |= _QUERY_OPTIONS["exhaust"]
-        if partial:
-            self.__query_flags |= _QUERY_OPTIONS["partial"]
+        self.__codec_options = collection.codec_options
+        self.__read_preference = collection.read_preference
 
-        # this is for passing network_timeout through if it's specified
-        # need to use kwargs as None is a legit value for network_timeout
-        self.__kwargs = kwargs
+        self.__query_flags = cursor_type
+        if self.__read_preference != ReadPreference.PRIMARY:
+            self.__query_flags |= _QUERY_OPTIONS["slave_okay"]
+        if no_cursor_timeout:
+            self.__query_flags |= _QUERY_OPTIONS["no_timeout"]
+        if allow_partial_results:
+            self.__query_flags |= _QUERY_OPTIONS["partial"]
+        if oplog_replay:
+            self.__query_flags |= _QUERY_OPTIONS["oplog_replay"]
 
     @property
     def collection(self):
         """The :class:`~pymongo.collection.Collection` that this
         :class:`Cursor` is iterating.
-
-        .. versionadded:: 1.1
         """
         return self.__collection
 
-    @property
-    def conn_id(self):
-        """The server/client/pool this cursor lives on.
-
-        Could be (host, port), -1, or None depending on what
-        client class executed the initial query or this cursor
-        being advanced at all.
-        """
-        return self.__connection_id
-
     @property
     def retrieved(self):
         """The number of documents retrieved so far.
@@ -228,7 +221,7 @@ class Cursor(object):
         """
         self.__data = deque()
         self.__id = None
-        self.__connection_id = None
+        self.__address = None
         self.__retrieved = 0
         self.__killed = False
 
@@ -245,16 +238,14 @@ class Cursor(object):
         return self._clone(True)
 
     def _clone(self, deepcopy=True):
+        """Internal clone helper."""
         clone = self._clone_base()
-        values_to_clone = ("spec", "fields", "skip", "limit", "max_time_ms",
-                           "comment", "max", "min",
-                           "snapshot", "ordering", "explain", "hint",
-                           "batch_size", "max_scan", "as_class", "slave_okay",
-                           "manipulate", "read_preference", "tag_sets",
-                           "secondary_acceptable_latency_ms",
-                           "must_use_master", "uuid_subtype", "compile_re",
-                           "query_flags", "kwargs")
-        data = dict((k, v) for k, v in self.__dict__.iteritems()
+        values_to_clone = ("spec", "projection", "skip", "limit",
+                           "max_time_ms", "comment", "max", "min",
+                           "ordering", "explain", "hint", "batch_size",
+                           "max_scan", "manipulate", "query_flags",
+                           "modifiers")
+        data = dict((k, v) for k, v in iteritems(self.__dict__)
                     if k.startswith('_Cursor__') and k[9:] in values_to_clone)
         if deepcopy:
             data = self._deepcopy(data)
@@ -276,11 +267,8 @@ class Cursor(object):
                 # to stop the server from sending more data.
                 self.__exhaust_mgr.sock.close()
             else:
-                connection = self.__collection.database.connection
-                if self.__connection_id is not None:
-                    connection.close_cursor(self.__id, self.__connection_id)
-                else:
-                    connection.close_cursor(self.__id)
+                self.__collection.database.client.close_cursor(self.__id,
+                                                               self.__address)
         if self.__exhaust and self.__exhaust_mgr:
             self.__exhaust_mgr.close()
         self.__killed = True
@@ -295,7 +283,7 @@ class Cursor(object):
     def __query_spec(self):
         """Get the spec to use for a query.
         """
-        operators = {}
+        operators = self.__modifiers.copy()
         if self.__ordering:
             operators["$orderby"] = self.__ordering
         if self.__explain:
@@ -304,8 +292,6 @@ class Cursor(object):
             operators["$hint"] = self.__hint
         if self.__comment:
             operators["$comment"] = self.__comment
-        if self.__snapshot:
-            operators["$snapshot"] = True
         if self.__max_scan:
             operators["$maxScan"] = self.__max_scan
         if self.__max_time_ms is not None:
@@ -314,47 +300,11 @@ class Cursor(object):
             operators["$max"] = self.__max
         if self.__min:
             operators["$min"] = self.__min
-        # Only set $readPreference if it's something other than
-        # PRIMARY to avoid problems with mongos versions that
-        # don't support read preferences.
-        if (self.__collection.database.connection.is_mongos and
-            self.__read_preference != ReadPreference.PRIMARY):
-
-            has_tags = self.__tag_sets and self.__tag_sets != [{}]
-
-            # For maximum backwards compatibility, don't set $readPreference
-            # for SECONDARY_PREFERRED unless tags are in use. Just rely on
-            # the slaveOkay bit (set automatically if read preference is not
-            # PRIMARY), which has the same behavior.
-            if (self.__read_preference != ReadPreference.SECONDARY_PREFERRED or
-                has_tags):
-
-                read_pref = {
-                    'mode': read_preferences.mongos_mode(self.__read_preference)
-                }
-                if has_tags:
-                    read_pref['tags'] = self.__tag_sets
-
-                operators['$readPreference'] = read_pref
 
         if operators:
             # Make a shallow copy so we can cleanly rewind or clone.
             spec = self.__spec.copy()
 
-            # Only commands that can be run on secondaries should have any
-            # operators added to the spec.  Command queries can be issued
-            # by db.command or calling find_one on $cmd directly
-            if self.collection.name == "$cmd":
-                # Don't change commands that can't be sent to secondaries
-                command_name = spec and spec.keys()[0].lower() or ""
-                if command_name not in secondary_ok_commands:
-                    return spec
-                elif command_name == 'mapreduce':
-                    # mapreduce shouldn't be changed if its not inline
-                    out = spec.get('out')
-                    if not isinstance(out, dict) or not out.get('inline'):
-                        return spec
-
             # White-listed commands must be wrapped in $query.
             if "$query" not in spec:
                 # $query has to come first
@@ -373,21 +323,12 @@ class Cursor(object):
         # Checking spec.keys()[0] covers the case that the spec
         # was passed as an instance of SON or OrderedDict.
         elif ("query" in self.__spec and
-              (len(self.__spec) == 1 or self.__spec.keys()[0] == "query")):
+              (len(self.__spec) == 1 or
+               next(iter(self.__spec)) == "query")):
             return SON({"$query": self.__spec})
 
         return self.__spec
 
-    def __query_options(self):
-        """Get the query options string to use for this query.
-        """
-        options = self.__query_flags
-        if (self.__slave_okay
-            or self.__read_preference != ReadPreference.PRIMARY
-        ):
-            options |= _QUERY_OPTIONS["slave_okay"]
-        return options
-
     def __check_okay_to_chain(self):
         """Check if it is okay to chain more options onto this cursor.
         """
@@ -404,12 +345,10 @@ class Cursor(object):
             raise TypeError("mask must be an int")
         self.__check_okay_to_chain()
 
-        if mask & _QUERY_OPTIONS["slave_okay"]:
-            self.__slave_okay = True
         if mask & _QUERY_OPTIONS["exhaust"]:
             if self.__limit:
                 raise InvalidOperation("Can't use limit and exhaust together.")
-            if self.__collection.database.connection.is_mongos:
+            if self.__collection.database.client.is_mongos:
                 raise InvalidOperation('Exhaust cursors are '
                                        'not supported by mongos')
             self.__exhaust = True
@@ -427,8 +366,6 @@ class Cursor(object):
             raise TypeError("mask must be an int")
         self.__check_okay_to_chain()
 
-        if mask & _QUERY_OPTIONS["slave_okay"]:
-            self.__slave_okay = False
         if mask & _QUERY_OPTIONS["exhaust"]:
             self.__exhaust = False
 
@@ -448,7 +385,7 @@ class Cursor(object):
 
         .. mongodoc:: limit
         """
-        if not isinstance(limit, (int, long)):
+        if not isinstance(limit, integer_types):
             raise TypeError("limit must be an integer")
         if self.__exhaust:
             raise InvalidOperation("Can't use limit and exhaust together.")
@@ -476,10 +413,8 @@ class Cursor(object):
 
         :Parameters:
           - `batch_size`: The size of each batch of results requested.
-
-        .. versionadded:: 1.9
         """
-        if not isinstance(batch_size, (int, long)):
+        if not isinstance(batch_size, integer_types):
             raise TypeError("batch_size must be an integer")
         if batch_size < 0:
             raise ValueError("batch_size must be >= 0")
@@ -500,7 +435,7 @@ class Cursor(object):
         :Parameters:
           - `skip`: the number of results to skip
         """
-        if not isinstance(skip, (int, long)):
+        if not isinstance(skip, integer_types):
             raise TypeError("skip must be an integer")
         if skip < 0:
             raise ValueError("skip must be >= 0")
@@ -522,7 +457,8 @@ class Cursor(object):
         :Parameters:
           - `max_time_ms`: the time limit after which the operation is aborted
         """
-        if not isinstance(max_time_ms, (int, long)) and max_time_ms is not None:
+        if (not isinstance(max_time_ms, integer_types)
+                and max_time_ms is not None):
             raise TypeError("max_time_ms must be an integer or None")
         self.__check_okay_to_chain()
 
@@ -585,7 +521,7 @@ class Cursor(object):
             self.__limit = limit
             return self
 
-        if isinstance(index, (int, long)):
+        if isinstance(index, integer_types):
             if index < 0:
                 raise IndexError("Cursor instances do not support negative"
                                  "indices")
@@ -607,10 +543,6 @@ class Cursor(object):
 
         :Parameters:
           - `max_scan`: the maximum number of documents to scan
-
-        .. note:: Requires server version **>= 1.5.1**
-
-        .. versionadded:: 1.7
         """
         self.__check_okay_to_chain()
         self.__max_scan = max_scan
@@ -706,13 +638,10 @@ class Cursor(object):
 
           collection.find({'field': 'value'}).hint('field_1').count()
 
-        With :class:`~pymongo.mongo_replica_set_client.MongoReplicaSetClient`
-        or :class:`~pymongo.master_slave_connection.MasterSlaveConnection`,
-        if `read_preference` is not
-        :attr:`pymongo.read_preferences.ReadPreference.PRIMARY` or
-        :attr:`pymongo.read_preferences.ReadPreference.PRIMARY_PREFERRED`, or
-        (deprecated) `slave_okay` is `True`, the count command will be sent to
-        a secondary or slave.
+        The :meth:`count` method obeys the
+        :attr:`~pymongo.collection.Collection.read_preference` of the
+        :class:`~pymongo.collection.Collection` instance on which
+        :meth:`~pymongo.collection.Collection.find` was called.
 
         :Parameters:
           - `with_limit_and_skip` (optional): take any :meth:`limit` or
@@ -722,53 +651,26 @@ class Cursor(object):
         .. note:: The `with_limit_and_skip` parameter requires server
            version **>= 1.1.4-**
 
-        .. note:: ``count`` ignores ``network_timeout``. For example, the
-          timeout is ignored in the following code::
-
-            collection.find({}, network_timeout=1).count()
-
         .. versionchanged:: 2.8
            The :meth:`~count` method now supports :meth:`~hint`.
-
-        .. versionadded:: 1.1.1
-           The `with_limit_and_skip` parameter.
-           :meth:`~pymongo.cursor.Cursor.__len__` was deprecated in favor of
-           calling :meth:`count` with `with_limit_and_skip` set to ``True``.
         """
-        if not isinstance(with_limit_and_skip, bool):
-            raise TypeError("with_limit_and_skip must be an instance of bool")
-        command = {"query": self.__spec, "fields": self.__fields}
-
-        command['read_preference'] = self.__read_preference
-        command['tag_sets'] = self.__tag_sets
-        command['secondary_acceptable_latency_ms'] = (
-            self.__secondary_acceptable_latency_ms)
-        command['slave_okay'] = self.__slave_okay
-        use_master = not self.__slave_okay and not self.__read_preference
-        command['_use_master'] = use_master
+        validate_boolean("with_limit_and_skip", with_limit_and_skip)
+        options = {"query": self.__spec}
         if self.__max_time_ms is not None:
-            command["maxTimeMS"] = self.__max_time_ms
+            options["maxTimeMS"] = self.__max_time_ms
         if self.__comment:
-            command['$comment'] = self.__comment
+            options["$comment"] = self.__comment
 
         if self.__hint is not None:
-            command['hint'] = self.__hint
+            options["hint"] = self.__hint
 
         if with_limit_and_skip:
             if self.__limit:
-                command["limit"] = self.__limit
+                options["limit"] = self.__limit
             if self.__skip:
-                command["skip"] = self.__skip
+                options["skip"] = self.__skip
 
-        database = self.__collection.database
-        r = database.command("count", self.__collection.name,
-                             allowable_errors=["ns missing"],
-                             uuid_subtype=self.__uuid_subtype,
-                             compile_re=self.__compile_re,
-                             **command)
-        if r.get("errmsg", "") == "ns missing":
-            return 0
-        return int(r["n"])
+        return self.__collection.count(**options)
 
     def distinct(self, key):
         """Get a list of distinct values for `key` among all documents
@@ -777,48 +679,25 @@ class Cursor(object):
         Raises :class:`TypeError` if `key` is not an instance of
         :class:`basestring` (:class:`str` in python 3).
 
-        With :class:`~pymongo.mongo_replica_set_client.MongoReplicaSetClient`
-        or :class:`~pymongo.master_slave_connection.MasterSlaveConnection`,
-        if `read_preference` is
-        not :attr:`pymongo.read_preferences.ReadPreference.PRIMARY` or
-        (deprecated) `slave_okay` is `True` the distinct command will be sent
-        to a secondary or slave.
+        The :meth:`distinct` method obeys the
+        :attr:`~pymongo.collection.Collection.read_preference` of the
+        :class:`~pymongo.collection.Collection` instance on which
+        :meth:`~pymongo.collection.Collection.find` was called.
 
         :Parameters:
           - `key`: name of key for which we want to get the distinct values
 
-        .. note:: Requires server version **>= 1.1.3+**
-
         .. seealso:: :meth:`pymongo.collection.Collection.distinct`
-
-        .. versionadded:: 1.2
         """
-        if not isinstance(key, basestring):
-            raise TypeError("key must be an instance "
-                            "of %s" % (basestring.__name__,))
-
-        options = {"key": key}
+        options = {}
         if self.__spec:
             options["query"] = self.__spec
-
-        options['read_preference'] = self.__read_preference
-        options['tag_sets'] = self.__tag_sets
-        options['secondary_acceptable_latency_ms'] = (
-            self.__secondary_acceptable_latency_ms)
-        options['slave_okay'] = self.__slave_okay
-        use_master = not self.__slave_okay and not self.__read_preference
-        options['_use_master'] = use_master
         if self.__max_time_ms is not None:
             options['maxTimeMS'] = self.__max_time_ms
         if self.__comment:
             options['$comment'] = self.__comment
 
-        database = self.__collection.database
-        return database.command("distinct",
-                                self.__collection.name,
-                                uuid_subtype=self.__uuid_subtype,
-                                compile_re=self.__compile_re,
-                                **options)["values"]
+        return self.__collection.distinct(key, **options)
 
     def explain(self):
         """Returns an explain plan record for this cursor.
@@ -831,7 +710,7 @@ class Cursor(object):
         # always use a hard limit for explains
         if c.__limit:
             c.__limit = -abs(c.__limit)
-        return c.next()
+        return next(c)
 
     def hint(self, index):
         """Adds a 'hint', telling Mongo the proper index to use for the query.
@@ -862,7 +741,7 @@ class Cursor(object):
             self.__hint = None
             return self
 
-        if isinstance(index, basestring):
+        if isinstance(index, string_type):
             self.__hint = index
         else:
             self.__hint = helpers._index_document(index)
@@ -909,31 +788,35 @@ class Cursor(object):
         self.__spec["$where"] = code
         return self
 
-    def __send_message(self, message):
-        """Send a query or getmore message and handles the response.
+    def __send_message(self, operation):
+        """Send a query or getmore operation and handles the response.
 
-        If message is ``None`` this is an exhaust cursor, which reads
+        If operation is ``None`` this is an exhaust cursor, which reads
         the next result batch off the exhaust socket instead of
         sending getMore messages to the server.
-        """
-        client = self.__collection.database.connection
 
-        if message:
-            kwargs = {"_must_use_master": self.__must_use_master}
-            kwargs["read_preference"] = self.__read_preference
-            kwargs["tag_sets"] = self.__tag_sets
-            kwargs["secondary_acceptable_latency_ms"] = (
-                self.__secondary_acceptable_latency_ms)
-            kwargs['exhaust'] = self.__exhaust
-            if self.__connection_id is not None:
-                kwargs["_connection_to_use"] = self.__connection_id
-            kwargs.update(self.__kwargs)
+        Can raise ConnectionFailure.
+        """
+        client = self.__collection.database.client
+
+        if operation:
+            kwargs = {
+                "read_preference": self.__read_preference,
+                "exhaust": self.__exhaust,
+            }
+            if self.__address is not None:
+                kwargs["address"] = self.__address
 
             try:
-                res = client._send_message_with_response(message, **kwargs)
-                self.__connection_id, (response, sock, pool) = res
+                response = client._send_message_with_response(operation,
+                                                              **kwargs)
+                self.__address = response.address
                 if self.__exhaust:
-                    self.__exhaust_mgr = _SocketManager(sock, pool)
+                    # 'response' is an ExhaustResponse.
+                    self.__exhaust_mgr = _SocketManager(response.socket_info,
+                                                        response.pool)
+
+                data = response.data
             except AutoReconnect:
                 # Don't try to send kill cursors on another socket
                 # or to another server. It can cause a _pinValue
@@ -944,22 +827,21 @@ class Cursor(object):
         else:
             # Exhaust cursor - no getMore message.
             try:
-                response = client._exhaust_next(self.__exhaust_mgr.sock)
-            except AutoReconnect:
-                self.__killed = True
-                self.__exhaust_mgr.error()
+                data = self.__exhaust_mgr.sock.receive_message(1, None)
+            except ConnectionFailure:
+                self.__die()
                 raise
 
         try:
-            response = helpers._unpack_response(response, self.__id,
-                                                self.__as_class,
-                                                self.__tz_aware,
-                                                self.__uuid_subtype,
-                                                self.__compile_re)
+            doc = helpers._unpack_response(response=data,
+                                           cursor_id=self.__id,
+                                           codec_options=self.__codec_options)
         except OperationFailure:
             self.__killed = True
+
             # Make sure exhaust socket is returned immediately, if necessary.
             self.__die()
+
             # If this is a tailable cursor the error is likely
             # due to capped collection roll over. Setting
             # self.__killed to True ensures Cursor.alive will be
@@ -967,25 +849,26 @@ class Cursor(object):
             if self.__query_flags & _QUERY_OPTIONS["tailable_cursor"]:
                 return
             raise
-        except AutoReconnect:
+        except NotMasterError:
             # Don't send kill cursors to another server after a "not master"
             # error. It's completely pointless.
             self.__killed = True
+
             # Make sure exhaust socket is returned immediately, if necessary.
             self.__die()
-            client.disconnect()
-            raise
 
-        self.__id = response["cursor_id"]
+            client._reset_server_and_request_check(self.__address)
+            raise
+        self.__id = doc["cursor_id"]
 
         # starting from doesn't get set on getmore's for tailable cursors
-        if not (self.__query_flags & _QUERY_OPTIONS["tailable_cursor"]):
-            assert response["starting_from"] == self.__retrieved, (
+        if not self.__query_flags & _QUERY_OPTIONS["tailable_cursor"]:
+            assert doc["starting_from"] == self.__retrieved, (
                 "Result batch started from %s, expected %s" % (
-                    response['starting_from'], self.__retrieved))
+                    doc['starting_from'], self.__retrieved))
 
-        self.__retrieved += response["number_returned"]
-        self.__data = deque(response["data"])
+        self.__retrieved += doc["number_returned"]
+        self.__data = deque(doc["data"])
 
         if self.__limit and self.__id and self.__limit <= self.__retrieved:
             self.__die()
@@ -1012,12 +895,14 @@ class Cursor(object):
                     ntoreturn = min(self.__limit, self.__batch_size)
                 else:
                     ntoreturn = self.__limit
-            self.__send_message(
-                message.query(self.__query_options(),
-                              self.__collection.full_name,
-                              self.__skip, ntoreturn,
-                              self.__query_spec(), self.__fields,
-                              self.__uuid_subtype))
+            self.__send_message(_Query(self.__query_flags,
+                                       self.__collection.full_name,
+                                       self.__skip,
+                                       ntoreturn,
+                                       self.__query_spec(),
+                                       self.__projection,
+                                       self.__codec_options,
+                                       self.__read_preference))
             if not self.__id:
                 self.__killed = True
         elif self.__id:  # Get More
@@ -1032,9 +917,9 @@ class Cursor(object):
             if self.__exhaust:
                 self.__send_message(None)
             else:
-                self.__send_message(
-                    message.get_more(self.__collection.full_name,
-                                     limit, self.__id))
+                self.__send_message(_GetMore(self.__collection.full_name,
+                                             limit,
+                                             self.__id))
 
         else:  # Cursor id is zero nothing else to return
             self.__killed = True
@@ -1049,8 +934,6 @@ class Cursor(object):
         <http://www.mongodb.org/display/DOCS/Tailable+Cursors>`_
         since they will stop iterating even though they *may* return more
         results in the future.
-
-        .. versionadded:: 1.5
         """
         return bool(len(self.__data) or (not self.__killed))
 
@@ -1066,22 +949,33 @@ class Cursor(object):
         """
         return self.__id
 
+    @property
+    def address(self):
+        """The (host, port) of the server used, or None.
+
+        .. versionchanged:: 3.0
+           Renamed from "conn_id".
+        """
+        return self.__address
+
     def __iter__(self):
         return self
 
-    def next(self):
+    def __next__(self):
         if self.__empty:
             raise StopIteration
-        db = self.__collection.database
+        _db = self.__collection.database
         if len(self.__data) or self._refresh():
             if self.__manipulate:
-                return db._fix_outgoing(self.__data.popleft(),
-                                        self.__collection)
+                return _db._fix_outgoing(self.__data.popleft(),
+                                         self.__collection)
             else:
                 return self.__data.popleft()
         else:
             raise StopIteration
 
+    next = __next__
+
     def __enter__(self):
         return self
 
@@ -1111,7 +1005,7 @@ class Cursor(object):
         if not hasattr(x, 'items'):
             y, is_list, iterator = [], True, enumerate(x)
         else:
-            y, is_list, iterator = {}, False, x.iteritems()
+            y, is_list, iterator = {}, False, iteritems(x)
 
         if memo is None:
             memo = {}
